@@ -5,14 +5,23 @@ import (
 	"encoding/binary"
 	"fmt"
 	"io"
+	"log"
 	"math/rand"
 	"net"
+	"net/http"
+	_ "net/http/pprof"
 	"strings"
 	"sync"
 	"testing"
 	"time"
 )
 
+func init() {
+	go func() {
+		log.Println(http.ListenAndServe("0.0.0.0:6060", nil))
+	}()
+}
+
 // setupServer starts new server listening on a random localhost port and
 // returns address of the server, function to stop the server, new client
 // connection to this server or an error.
@@ -188,6 +197,19 @@ func TestCloseThenOpen(t *testing.T) {
 	}
 }
 
+func TestSessionDoubleClose(t *testing.T) {
+	_, stop, cli, err := setupServer(t)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer stop()
+	session, _ := Client(cli, nil)
+	session.Close()
+	if err := session.Close(); err == nil {
+		t.Fatal("session double close doesn't return error")
+	}
+}
+
 func TestStreamDoubleClose(t *testing.T) {
 	_, stop, cli, err := setupServer(t)
 	if err != nil {
@@ -198,7 +220,7 @@ func TestStreamDoubleClose(t *testing.T) {
 	stream, _ := session.OpenStream()
 	stream.Close()
 	if err := stream.Close(); err == nil {
-		t.Log("double close doesn't return error")
+		t.Fatal("stream double close doesn't return error")
 	}
 	session.Close()
 }
@@ -304,6 +326,44 @@ func TestKeepAliveTimeout(t *testing.T) {
 	}
 }
 
+type blockWriteConn struct {
+	net.Conn
+}
+
+func (c *blockWriteConn) Write(b []byte) (n int, err error) {
+	forever := time.Hour * 24
+	time.Sleep(forever)
+	return c.Conn.Write(b)
+}
+
+func TestKeepAliveBlockWriteTimeout(t *testing.T) {
+	ln, err := net.Listen("tcp", "localhost:0")
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer ln.Close()
+	go func() {
+		ln.Accept()
+	}()
+
+	cli, err := net.Dial("tcp", ln.Addr().String())
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer cli.Close()
+	//when writeFrame block, keepalive in old version never timeout
+	blockWriteCli := &blockWriteConn{cli}
+
+	config := DefaultConfig()
+	config.KeepAliveInterval = time.Second
+	config.KeepAliveTimeout = 2 * time.Second
+	session, _ := Client(blockWriteCli, config)
+	time.Sleep(3 * time.Second)
+	if !session.IsClosed() {
+		t.Fatal("keepalive-timeout failed")
+	}
+}
+
 func TestServerEcho(t *testing.T) {
 	ln, err := net.Listen("tcp", "localhost:0")
 	if err != nil {
@@ -541,6 +601,95 @@ func TestRandomFrame(t *testing.T) {
 
 	session.conn.Write(buf)
 	cli.Close()
+
+	// writeFrame after die
+	cli, err = net.Dial("tcp", addr)
+	if err != nil {
+		t.Fatal(err)
+	}
+	session, _ = Client(cli, nil)
+	//close first
+	session.Close()
+	for i := 0; i < 100; i++ {
+		f := newFrame(byte(rand.Uint32()), rand.Uint32())
+		session.writeFrame(f)
+	}
+}
+
+func TestWriteFrameInternal(t *testing.T) {
+	addr, stop, cli, err := setupServer(t)
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer stop()
+	// pure random
+	session, _ := Client(cli, nil)
+	for i := 0; i < 100; i++ {
+		rnd := make([]byte, rand.Uint32()%1024)
+		io.ReadFull(crand.Reader, rnd)
+		session.conn.Write(rnd)
+	}
+	cli.Close()
+
+	// writeFrame after die
+	cli, err = net.Dial("tcp", addr)
+	if err != nil {
+		t.Fatal(err)
+	}
+	session, _ = Client(cli, nil)
+	//close first
+	session.Close()
+	for i := 0; i < 100; i++ {
+		f := newFrame(byte(rand.Uint32()), rand.Uint32())
+		session.writeFrameInternal(f, time.After(session.config.KeepAliveTimeout))
+	}
+
+	// random cmds
+	cli, err = net.Dial("tcp", addr)
+	if err != nil {
+		t.Fatal(err)
+	}
+	allcmds := []byte{cmdSYN, cmdFIN, cmdPSH, cmdNOP}
+	session, _ = Client(cli, nil)
+	for i := 0; i < 100; i++ {
+		f := newFrame(allcmds[rand.Int()%len(allcmds)], rand.Uint32())
+		session.writeFrameInternal(f, time.After(session.config.KeepAliveTimeout))
+	}
+	//deadline occur
+	{
+		c := make(chan time.Time)
+		close(c)
+		f := newFrame(allcmds[rand.Int()%len(allcmds)], rand.Uint32())
+		_, err := session.writeFrameInternal(f, c)
+		if !strings.Contains(err.Error(), "timeout") {
+			t.Fatal("write frame with deadline failed", err)
+		}
+	}
+	cli.Close()
+
+	{
+		cli, err = net.Dial("tcp", addr)
+		if err != nil {
+			t.Fatal(err)
+		}
+		config := DefaultConfig()
+		config.KeepAliveInterval = time.Second
+		config.KeepAliveTimeout = 2 * time.Second
+		session, _ = Client(&blockWriteConn{cli}, config)
+		f := newFrame(byte(rand.Uint32()), rand.Uint32())
+		c := make(chan time.Time)
+		go func() {
+			//die first, deadline second, better for coverage
+			time.Sleep(time.Second)
+			session.Close()
+			time.Sleep(time.Second)
+			close(c)
+		}()
+		_, err = session.writeFrameInternal(f, c)
+		if !strings.Contains(err.Error(), "closed pipe") {
+			t.Fatal("write frame with to closed conn failed", err)
+		}
+	}
 }
 
 func TestReadDeadline(t *testing.T) {
@@ -561,7 +710,7 @@ func TestReadDeadline(t *testing.T) {
 		}
 	}
 	if readErr != nil {
-		if !strings.Contains(readErr.Error(), "i/o timeout") {
+		if !strings.Contains(readErr.Error(), "timeout") {
 			t.Fatalf("Wrong error: %v", readErr)
 		}
 	} else {
@@ -583,7 +732,7 @@ func TestWriteDeadline(t *testing.T) {
 	for {
 		stream.SetWriteDeadline(time.Now().Add(-1 * time.Minute))
 		if _, writeErr = stream.Write(buf); writeErr != nil {
-			if !strings.Contains(writeErr.Error(), "i/o timeout") {
+			if !strings.Contains(writeErr.Error(), "timeout") {
 				t.Fatalf("Wrong error: %v", writeErr)
 			}
 			break
@@ -693,6 +842,7 @@ func bench(b *testing.B, rd io.Reader, wr io.Writer) {
 	buf2 := make([]byte, 128*1024)
 	b.SetBytes(128 * 1024)
 	b.ResetTimer()
+	b.ReportAllocs()
 
 	var wg sync.WaitGroup
 	wg.Add(1)
